/*
 *  This code was written by Yufei Yuan <yuanyufei@hotmail.com>
 */

#include "libQccPack.h"
#include "libQccPackLCM.h"

static int QccWAVlcmBitBufferPutBits(QccBitBuffer *bit_buffer, int count, int k)
{
	int i, bit_value, mask;
	int return_value;

	if(bit_buffer == NULL) return(0);

	if(bit_buffer->type != QCCBITBUFFER_OUTPUT) {
		QccErrorAddMessage("(QccWAVlcmBitBufferPutBits): Not an output buffer");
		goto QccError;
	}

	mask = 1 << (k - 1);

	for(i = 0; i < k; i++) {
		if (count & mask) bit_value = 1;
		else bit_value = 0;
		mask >>= 1;

		if(QccBitBufferPutBit(bit_buffer, bit_value)) {
			QccErrorAddMessage("(QccWAVlcmBitBufferPutBits): Error calling QccBitBufferPutBit()");
			goto QccError;
		}
	}

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	return(return_value);
}

static int QccWAVlcmBitBufferGetBits(QccBitBuffer *bit_buffer, int *count, int k)
{
	int i, bit_value;
	int return_value;

	if(bit_buffer == NULL)
		return(0);

	if(bit_buffer->type != QCCBITBUFFER_INPUT) {
		QccErrorAddMessage("(QccWAVlcmBitBufferPutBits): Not an input buffer");
		goto QccError;
	}

	for(i = 0, *count = 0; i < k; i++) {
		if(QccBitBufferGetBit(bit_buffer, &bit_value)) {
			QccErrorAddMessage("(QccWAVlcmBitBufferGetBits): Error calling QccBitBufferGetBit()");
			goto QccError;
		}
		*count = (*count << 1) | bit_value;
	}

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	return(return_value);
}

int QccWAVlcmAdaptiveGolombEncode(QccBitBuffer *output_buffer, const int *symbols, 
	int num_symbols, QccChannel *sign, int Kmax)
{
	int symbol, count, i, k, limit, value;
	int return_value;

	if(output_buffer == NULL) return(0);
	if(symbols == NULL) return(0);

	for(count = 0, limit = 1, i = 0, k = 0, symbol = 0; symbol < num_symbols; symbol++) {
		if(symbols[symbol] == QCCWAVLCM_REPEAT) {
			if(k == 0) {
				value = 1;
				if(QccBitBufferPutBit(output_buffer, value)) {
					QccErrorAddMessage("(QccWAVlcmAdaptiveGolombEncode): Error calling QccBitBufferPutBit()");
					goto QccError;
				}
				k++; limit = (int)pow(2., (double)k);
				count = 0;
			}
			else {
				count++;
				if(count == limit) {
					value = 1;
					if(QccBitBufferPutBit(output_buffer, value)) {
						QccErrorAddMessage("(QccWAVlcmAdaptiveGolombEncode): Error calling QccBitBufferPutBit()");
						goto QccError;
					}
					if(k < Kmax) {k++; limit = (int)pow(2., (double)k);}
					count = 0;
				}
			}
		}
		else if(symbols[symbol] == QCCWAVLCM_STOP) {
			value = 0;
			if(QccBitBufferPutBit(output_buffer, value)) {
				QccErrorAddMessage("(QccWAVlcmAdaptiveGolombEncode): Error calling QccBitBufferPutBit()");
				goto QccError;
			}
			if(QccWAVlcmBitBufferPutBits(output_buffer, count, k)) {
				QccErrorAddMessage("(QccWAVlcmAdaptiveGolombEncode): Error calling QccWAVlcmBitBufferPutBits()");
				goto QccError;
			}
			if(QccBitBufferPutBit(output_buffer, sign->channel_symbols[i++])) {
				QccErrorAddMessage("(QccWAVlcmAdaptiveGolombEncode): Error calling QccWAVlcmBitBufferPutBit()");
				goto QccError;
			}
			count = 0;
			if(k != 0) {k--; limit = (int)pow(2., (double)k);}
		}
		else {
			QccErrorAddMessage("(QccWAVlcmAdaptiveGolombEncode): Error reading symbols.");
			goto QccError;
		}
	}
	if(QccBitBufferPutBit(output_buffer, 1)) {
		QccErrorAddMessage("(QccWAVlcmAdaptiveGolombEncode): Error calling QccBitBufferPutBit()");
		goto QccError;
	}

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	return(return_value);
}

int QccWAVlcmAdaptiveGolombDecode(QccBitBuffer *input_buffer, int *symbols, 
	int num_symbols, QccChannel *sign, int Kmax)
{
	int i, symbol, count, k, limit, value;
	int return_value;

	if(input_buffer == NULL) return(0);
	if(symbols == NULL) return(0);

	count = 0; k = 0; symbol = 0;
	while(symbol < num_symbols) {
		if(QccBitBufferGetBit(input_buffer, &value)) {
			QccErrorAddMessage("(QccWAVlcmAdaptiveGolombDecode): Error calling QccBitBufferGetBit()");
			goto QccError;
		}
		if(value == 1) {
			limit = (int)pow(2., (double)k);
			for(i = 0; i < limit; i++) {
				symbols[symbol] = QCCWAVLCM_REPEAT; 
				symbol++;
				if(symbol >= num_symbols) break;
			}
			/* read in one redundant bit and discard it */
			if(i == (limit - 1) && symbol == num_symbols) {
				if(QccBitBufferGetBit(input_buffer, &value)) {
					QccErrorAddMessage("(QccWAVlcmAdaptiveGolombDecode): Error calling QccBitBufferGetBit()");
					goto QccError;
				}
			}
			if(k < Kmax) k++;
			count = 0;
		}
		else {
			if(QccWAVlcmBitBufferGetBits(input_buffer, &count, k)) {
				QccErrorAddMessage("(QccWAVlcmAdaptiveGolombDecode): Error calling QccWAVlcmBitBufferGetBits()");
				goto QccError;
			}
			for(i = 0; i < count; i++) {
				symbols[symbol] = QCCWAVLCM_REPEAT; 
				symbol++;
			}
			symbols[symbol++] = QCCWAVLCM_STOP;
			if(QccBitBufferGetBit(input_buffer, &value)) {
				QccErrorAddMessage("(QccWAVlcmAdaptiveGolombDecode): Error calling QccBitBufferGetBit()");
				goto QccError;
			}
			if(QccChannelAddSymbolToChannel(sign, value)) {
				QccErrorAddMessage("(QccWAVlcmAdaptiveGolombDecode): Error calling QccChannelAddSymbolToChannel()");
				goto QccError;
			}
			/* read in one redundant bit and discard it */
			if(symbol == num_symbols) {
				if(QccBitBufferGetBit(input_buffer, &value)) {
					QccErrorAddMessage("(QccWAVlcmAdaptiveGolombDecode): Error calling QccBitBufferGetBit()");
					goto QccError;
				}
			}
			if(k > 0) k--;
		}
	}

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	return(return_value);
}

void QccWAVlcmListFreeNode(QccListNode *node)
{
	if(node == NULL) return;

	if(node->value != NULL) {
		/* free(node->value); */
		node->value = NULL;
	}
	free(node);
}

void QccWAVlcmListFree(QccList *list)
{
	QccListNode *current;
	QccListNode *next;

	if(list == NULL) return;

	current = list->start;
	while(current != NULL) {
		next = current->next;
		QccWAVlcmListFreeNode(current);
		current = next;
	}
	list->start = NULL;
	list->end = NULL;
}

QccListNode *QccWAVlcmListCreateNode(int value_size, void *value)
{
	QccListNode *new_node = NULL;

	if((new_node = (QccListNode *)malloc(sizeof(QccListNode))) == NULL)
	{
		QccErrorAddMessage("(QccWAVlcmListCreateNode): Error allocating memory");
		return(NULL);
	}

	new_node->previous = NULL;
	new_node->next = NULL;
	new_node->value_size = value_size;

	if(value != NULL) new_node->value = value;
	else {
		QccErrorAddMessage("(QccWAVlcmListCreateNode): Node should not point to NULL");
		return(NULL);  
	}

	return(new_node);
}

void QccWAVlcmHeader(void)
{
	QccString header;

	QccStringSprintf(header, "%s\n\nQccPackLCM Module Version %d.%d %s,\n%s\n\nfor:", QCCWAVLCM_LICENSE,
		QCCWAVLCM_MAJORVERSION, QCCWAVLCM_MINORVERSION, QCCWAVLCM_DATE, QCCWAVLCM_COPYRIGHT);
	QccSetUserHeader(header);
}

int QccWAVlcmEncodeHeader(QccBitBuffer *buffer, int num_levels, int num_rows, int num_cols,
	double image_mean, int max_coefficient_bits)
{
	int return_value;

	if(QccBitBufferPutChar(buffer, (unsigned char)num_levels)) {
		QccErrorAddMessage("(QccWAVlcmEncodeHeader): Error calling QccBitBufferPuChar()");
		goto QccError;
	}
	if(QccBitBufferPutInt(buffer, num_rows)) {
		QccErrorAddMessage("(QccWAVlcmEncodeHeader): Error calling QccBitBufferPutInt()");
		goto QccError;
	}
	if(QccBitBufferPutInt(buffer, num_cols)) {
		QccErrorAddMessage("(QccWAVlcmEncodeHeader): Error calling QccBitBufferPutInt()");
		goto QccError;
	}
	if(QccBitBufferPutDouble(buffer, image_mean)) {
		QccErrorAddMessage("(QccWAVlcmEncodeHeader): Error calling QccBitBufferPutDouble()");
		goto QccError;
	}
	if(QccBitBufferPutChar(buffer, (unsigned char)max_coefficient_bits)) {
		QccErrorAddMessage("(QccWAVlcmEncodeHeader): Error calling QccBitBufferPutChar()");
		goto QccError;
	}

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	return(return_value);
}

static int QccWAVlcmInputOutput(QccBitBuffer *buffer, int *symbol, int method, int target_bit_cnt)
{
	int return_value = 0;

	if(method == QCCWAVLCM_ENCODE) {
		if(QccBitBufferPutBit(buffer, *symbol)) {
			QccErrorAddMessage("(QccWAVlcmInputOutput): Error calling QccBitBufferPutBit()");
			goto QccError;
		}
		if(buffer->bit_cnt >= target_bit_cnt) {
			QccBitBufferFlush(buffer);
			goto QccError;
		}
	}
	else {
		if(QccBitBufferGetBit(buffer, symbol)) {
			QccErrorAddMessage("(QccWAVlcmInputOutput): Error calling QccBitBufferGetBit()");
			goto QccError;
		}
	}

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	return(return_value);
}

static int QccWAVlcmInputOutputLLThreshold(QccBitBuffer *buffer, QccWAVSubbandPyramid *subband_pyramid, 
	unsigned char *LL_max_bits, int method)
{
	QccMatrix LL_subband_matrix;
	int i, j, num_rows, num_cols;
	int max, return_value;

	if(QccWAVSubbandPyramidSubbandResolution(subband_pyramid, 0, &num_rows, &num_cols)) {
		QccErrorAddMessage("(QccWAVlcmInputOutputLLMean): Error calling QccWAVSubbandPyramidSubbandResolution()");
		goto QccError;    
	}

	if(method == QCCWAVLCM_ENCODE) {
		if((LL_subband_matrix = QccMatrixAlloc(num_rows, num_cols)) == NULL) {
			QccErrorAddMessage("(QccWAVlcmInputOutputLLMean): Error calling QccMatrixAlloc()");
			goto QccError;    
		}

		for(i = 0; i < num_rows; i++) {
			for(j = 0; j < num_cols; j++) LL_subband_matrix[i][j] = subband_pyramid->matrix[i][j];
		}
		max = (int)QccMatrixMaxValue(LL_subband_matrix, num_rows, num_cols);
		*LL_max_bits = (unsigned char)floor(QccMathLog2(max));

		if(QccBitBufferPutChar(buffer, *LL_max_bits)) {
			QccErrorAddMessage("(QccWAVlcmInputOutputLLMean): Error calling QccBitBufferPutChar()");
			goto QccError;
		}
	}
	else {
		if(QccBitBufferGetChar(buffer, LL_max_bits)) {
			QccErrorAddMessage("(QccWAVlcmInputOutputLLMean): Error calling QccBitBufferPutChar()");
			goto QccError;
		}
	}

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	if(method == QCCWAVLCM_ENCODE) QccMatrixFree(LL_subband_matrix, num_rows);
	return(return_value);
}

static int QccWAVlcmInputOutputLLBitplane(QccBitBuffer *buffer, QccWAVSubbandPyramid *subband_pyramid, 
	QccWAVlcmCoefficientBlock **subband_lcm_pyramid, int **sign_array, double threshold, 
	int LL_max_bits, int target_bit_cnt, int method)
{
	int i, j, num_rows, num_cols, bit, return_value;
	double LL_threshold;

	if(QccWAVSubbandPyramidSubbandResolution(subband_pyramid, 0, &num_rows, &num_cols)) {
		QccErrorAddMessage("(QccWAVlcmInputOutputLLBitplane): Error calling QccWAVSubbandPyramidSubbandResolution()");
		goto QccError;    
	}

	LL_threshold = pow((double)2, (double)LL_max_bits);

	if(method == QCCWAVLCM_ENCODE) {
		if(threshold <= LL_threshold) {
			if(threshold == LL_threshold) {
				for(i = 0; i < num_rows; i++) {
					for(j = 0; j < num_cols; j++) {
						if(QccWAVlcmInputOutput(buffer, &sign_array[i][j], QCCWAVLCM_ENCODE, target_bit_cnt)) {
							QccErrorAddMessage("(QccWAVlcmInputOutputLLBitplane): Error calling QccWAVlcmInputOutput()");
							return(2);    
						}
					}
				}
			}

			for(i = 0; i < num_rows; i++) {
				for(j = 0; j < num_cols; j++) {
					bit = subband_lcm_pyramid[i][j].value >= threshold;
					subband_lcm_pyramid[i][j].significant = bit;
					if(bit) subband_lcm_pyramid[i][j].value -= threshold;                
					if(QccWAVlcmInputOutput(buffer, &bit, QCCWAVLCM_ENCODE, target_bit_cnt)) {
						QccErrorAddMessage("(QccWAVlcmInputOutputLLBitplane): Error calling QccWAVlcmInputOutput()");
						return(2);    
					}
				}
			}
		}
	}
	else {
		if(threshold <= LL_threshold) {
			if(threshold == LL_threshold) {
				for(i = 0; i < num_rows; i++) {
					for(j = 0; j < num_cols; j++) {
						if(QccWAVlcmInputOutput(buffer, &sign_array[i][j], QCCWAVLCM_DECODE, target_bit_cnt)) {
							QccErrorAddMessage("(QccWAVlcmInputOutputLLBitplane): Error calling QccWAVlcmInputOutput()");
							return(2);
						}
						subband_lcm_pyramid[i][j].value = threshold;                        
					}
				}
			}

			for(i = 0; i < num_rows; i++) {
				for(j = 0; j < num_cols; j++) {
					if(QccWAVlcmInputOutput(buffer, &bit, QCCWAVLCM_DECODE, target_bit_cnt)) {
						QccErrorAddMessage("(QccWAVlcmInputOutputLLBitplane): Error calling QccWAVlcmInputOutput()");
						return(2);    
					}
					subband_lcm_pyramid[i][j].significant = bit;                    
					subband_lcm_pyramid[i][j].value += (bit) ? (threshold / 2) : (-threshold / 2);
				}
			}
		}
	}

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	return(return_value);
}

static int QccWAVlcmEncodeDWT(QccWAVSubbandPyramid *image_subband_pyramid, int **sign_array,
	const QccIMGImageComponent *image, int num_levels, double *image_mean,
	int *max_coefficient_bits, QccWAVSubbandPyramid *mask_subband_pyramid,
	const QccIMGImageComponent *mask, const QccWAVWavelet *wavelet,
	const QccWAVPerceptualWeights *perceptual_weights)
{
	/*double coefficient_magnitude;*/
	double max_coefficient = -MAXFLOAT;
	int row, col;
	int return_value;

	if(QccMatrixCopy(image_subband_pyramid->matrix, image->image, image->num_rows, 
		image->num_cols)) {
		QccErrorAddMessage("(QccWAVlcmEncodeDWT): Error calling QccMatrixCopy()");
		goto QccError;
	}

	if(mask != NULL) {
		if(QccMatrixCopy(mask_subband_pyramid->matrix, mask->image, mask->num_rows, 
			mask->num_cols)) {
			QccErrorAddMessage("(QccWAVlcmEncodeDWT): Error calling QccMatrixCopy()");
			goto QccError;
		}

		for(row = 0; row < image_subband_pyramid->num_rows; row++) {
			for (col = 0; col < image_subband_pyramid->num_cols; col++) {
				if (QccAlphaTransparent(mask_subband_pyramid->matrix[row][col]))
					image_subband_pyramid->matrix[row][col] = 0;
			}
		}
	}

	if(QccWAVSubbandPyramidSubtractMean(image_subband_pyramid, image_mean, NULL)) {
		QccErrorAddMessage("(QccWAVlcmEncodeDWT): Error calling QccWAVSubbandPyramidSubtractMean()");
		goto QccError;
	}

	if(mask != NULL) {
		if(QccWAVSubbandPyramidShapeAdaptiveDWT(image_subband_pyramid, mask_subband_pyramid,
			num_levels, wavelet)) {
			QccErrorAddMessage("(QccWAVlcmEncodeDWT): Error calling QccWAVSubbandPyramidShapeAdaptiveDWT()");
			goto QccError;
		}
	}
	else if(QccWAVSubbandPyramidDWT(image_subband_pyramid, num_levels, wavelet)) {
		QccErrorAddMessage("(QccWAVlcmEncodeDWT): Error calling QccWAVSubbandPyramidDWT()");
		goto QccError;
	}

	if(perceptual_weights != NULL) {
		if(QccWAVPerceptualWeightsApply(image_subband_pyramid, perceptual_weights)) {
			QccErrorAddMessage("(QccWAVlcmEncodeDWT): Error calling QccWAVPerceptualWeightsApply()");
			goto QccError;
		}
	}

	for(row = 0; row < image_subband_pyramid->num_rows; row++) {
		for (col = 0; col < image_subband_pyramid->num_cols; col++) {
			image_subband_pyramid->matrix[row][col] >= 0 ? (sign_array[row][col] = 0) : (sign_array[row][col] = 1);
			image_subband_pyramid->matrix[row][col] = fabs(image_subband_pyramid->matrix[row][col])	;
			if(image_subband_pyramid->matrix[row][col] > max_coefficient) max_coefficient = image_subband_pyramid->matrix[row][col];
		}
	}

	*max_coefficient_bits = (int)floor(QccMathLog2(max_coefficient));

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	return(return_value);
}

static int QccWAVlcmDecodeInverseDWT(QccWAVSubbandPyramid *image_subband_pyramid,
	QccWAVSubbandPyramid *mask_subband_pyramid, int **sign_array, QccIMGImageComponent *image,
	double image_mean, const QccWAVWavelet *wavelet, 
	const QccWAVPerceptualWeights *perceptual_weights) 
{
	int row, col;
	int return_value;

	for(row = 0; row < image_subband_pyramid->num_rows; row++)
		for(col = 0; col < image_subband_pyramid->num_cols; col++)
			if(sign_array[row][col]) image_subband_pyramid->matrix[row][col] *= -1;

	if(perceptual_weights != NULL) {
		if(QccWAVPerceptualWeightsRemove(image_subband_pyramid, perceptual_weights)) {
			QccErrorAddMessage("(QccWAVlcmDecodeInverseDWT): Error calling QccWAVPerceptualWeightsRemove()");
			goto QccError;
		}
	}

	if(mask_subband_pyramid != NULL) {
		if(QccWAVSubbandPyramidInverseShapeAdaptiveDWT(image_subband_pyramid,
			mask_subband_pyramid, wavelet)) {
				QccErrorAddMessage("(QccWAVlcmDecodeInverseDWT): Error calling QccWAVSubbandPyramidInverseShapeAdaptiveDWT()");
				goto QccError;
		}
	}
	else if(QccWAVSubbandPyramidInverseDWT(image_subband_pyramid, wavelet)) {
		QccErrorAddMessage("(QccWAVlcmDecodeInverseDWT): Error calling QccWAVSubbandPyramidInverseDWT()");
		goto QccError;
	}

	if(QccWAVSubbandPyramidAddMean(image_subband_pyramid, image_mean)) {
		QccErrorAddMessage("(QccWAVlcmDecodeInverseDWT): Error calling QccWAVSubbandPyramidAddMean()");
		goto QccError;
	}

	if(mask_subband_pyramid != NULL) {
		for(row = 0; row < image_subband_pyramid->num_rows; row++) {
			for(col = 0; col < image_subband_pyramid->num_cols; col++) {
				if(QccAlphaTransparent(mask_subband_pyramid->matrix[row][col]))
					image_subband_pyramid->matrix[row][col] = 0;
			}
		}
	}

	if(QccMatrixCopy(image->image, image_subband_pyramid->matrix, image->num_rows, 
		image->num_cols)) {
			QccErrorAddMessage("(QccWAVlcmDecodeInverseDWT): Error calling QccMatrixCopy()");
			goto QccError;
	}

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	return(return_value);
}

static void QccWAVlcmMacroBlockInit(int *macro_block_size, int *num_blocks, int num_levels, int row, int column, 
	int *blocks_per_row, int *blocks_per_col)
{
	int i, j;

	for(i = 0; i < num_levels; i++) macro_block_size[i] = 1 << (QCCWAVLCM_BLOCK_EXP - i - 1);
	/* blocks_per_row = (int)ceil((double)row / (double)(1 << QCCWAVLCM_BLOCK_EXP)); */
	i = 1 << QCCWAVLCM_BLOCK_EXP;
	j = row / i;
	*blocks_per_row = (row > i * j) ? (j + 1) : j;
	/*blocks_per_col = (int)ceil((double)column / (double)(1 << QCCWAVLCM_BLOCK_EXP)); */
	j = column / i;
	*blocks_per_col = (column > i * j) ? (j + 1) : j;

	*num_blocks = (*blocks_per_row) * (*blocks_per_col);
}

static int QccWAVlcmTagInitialize(QccWAVSubbandPyramid *subband_pyramid, const QccWAVSubbandPyramid *mask_subband_pyramid, 
	int **sign_array, int num_levels, QccWAVlcmCoefficientBlock **subband_lcm_pyramid)
{
	int return_value;
	int subband_num_rows, subband_num_cols;
	int row_offset, col_offset;
	int parent_subband_rows, parent_subband_cols;
	int parent_row_offset, parent_col_offset;
	int row, col, num_subbands, i, j, k, jj, kk, count;
	QccWAVlcmCoefficientBlock coefficient_block, **temp;
	int transparent;

	num_subbands = QccWAVSubbandPyramidNumLevelsToNumSubbands(num_levels);
	for(i = 0; i < num_subbands; i++) {
		if(QccWAVSubbandPyramidSubbandResolution(subband_pyramid, i, &subband_num_rows, &subband_num_cols)) {
			QccErrorAddMessage("(QccWAVlcmTagInitialize): Error calling QccWAVSubbandPyramidSubbandResolution()");
			goto QccError;
		}
		if(QccWAVSubbandPyramidSubbandOffsets(subband_pyramid, i, &row_offset, &col_offset)) {
			QccErrorAddMessage("(QccWAVlcmTagInitialize): Error calling QccWAVSubbandPyramidSubbandOffsets()");
			goto QccError;
		}

		if(i < 4) {
			if(i) {
				if(QccWAVSubbandPyramidSubbandResolution(subband_pyramid, 0, &parent_subband_rows, &parent_subband_cols)) {
					QccErrorAddMessage("(QccWAVlcmTagInitialize): Error calling QccWAVSubbandPyramidSubbandResolution()");
					goto QccError;
				}
				if(QccWAVSubbandPyramidSubbandOffsets(subband_pyramid, 0, &parent_row_offset, &parent_col_offset)) {
					QccErrorAddMessage("(QccWAVlcmTagInitialize): Error calling QccWAVSubbandPyramidSubbandOffsets()");
					goto QccError;
				}
			}
		}
		else {
			if(QccWAVSubbandPyramidSubbandResolution(subband_pyramid, i - 3, &parent_subband_rows, &parent_subband_cols)) {
				QccErrorAddMessage("(QccWAVlcmTagInitialize): Error calling QccWAVSubbandPyramidSubbandResolution()");
				goto QccError;
			}
			if(QccWAVSubbandPyramidSubbandOffsets(subband_pyramid, i - 3, &parent_row_offset, &parent_col_offset)) {
				QccErrorAddMessage("(QccWAVlcmTagInitialize): Error calling QccWAVSubbandPyramidSubbandOffsets()");
			}
		}

		for(row = 0; row < subband_num_rows; row++) {
			for(col = 0; col < subband_num_cols; col++) {
				coefficient_block.widthx = row + row_offset;
				coefficient_block.heighty = col + col_offset;

				if(!i) coefficient_block.parent = NULL;
				else {
					if(mask_subband_pyramid == NULL) transparent = 0;
					else {
						if(i < 4) transparent = QccAlphaTransparent(mask_subband_pyramid->matrix
							[row + parent_row_offset][col + col_offset]);
						else transparent = QccAlphaTransparent(mask_subband_pyramid->matrix
							[(row / 2) + parent_row_offset][(col / 2) + col_offset]);
					}

					if(i < 4) {
						if(transparent) coefficient_block.parent = NULL;
						else coefficient_block.parent = (void *)&subband_lcm_pyramid[parent_row_offset + row][parent_col_offset + col];
					}
					else {
						if(transparent) coefficient_block.parent = NULL;
						else coefficient_block.parent = (void *)&subband_lcm_pyramid[parent_row_offset + (row / 2)][parent_col_offset + (col / 2)];
					}
				}

				for(j = -1, count = 0; j < 2; j++) {
					for(k = -1; k < 2; k++) {
						if(j == 0 && k == 0) continue;
						jj = row + j; kk = col + k;
						if(jj >= 0 && jj < subband_num_rows && kk >= 0 && kk < subband_num_cols) {
							if(mask_subband_pyramid == NULL) transparent = 0;
							else transparent = QccAlphaTransparent(mask_subband_pyramid->matrix[jj + row_offset][kk + col_offset]);
							if(!transparent)count++;
						}
					}
				}
				coefficient_block.num_neighbors = count;
				if(count == 0) coefficient_block.neighbors = NULL;
				else {
					coefficient_block.neighbors = malloc(sizeof(QccWAVlcmCoefficientBlock *) * count);
					/* temp = (QccWAVlcmCoefficientBlock **)coefficient_block.neighbors; */
					temp = coefficient_block.neighbors;
					for(j = -1, count = 0; j < 2; j++) {
						for(k = -1; k < 2; k++) {
							if(j == 0 && k == 0) continue;
							jj = row + j; kk = col + k;
							if(jj >= 0 && jj < subband_num_rows && kk >= 0 && kk < subband_num_cols) {
								if(mask_subband_pyramid == NULL) transparent = 0;
								else transparent = QccAlphaTransparent(mask_subband_pyramid->matrix[jj + row_offset][kk + col_offset]);
								if(!transparent) temp[count++] = &subband_lcm_pyramid[row_offset + jj][col_offset + kk];
							}
						}
					}
				}
				/* coefficient_block.sign = sign_array[coefficient_block.widthx][coefficient_block.heighty]; */
				coefficient_block.value = fabs(subband_pyramid->matrix[coefficient_block.widthx][coefficient_block.heighty]);
				coefficient_block.significant = 0;
				memcpy(&subband_lcm_pyramid[coefficient_block.widthx][coefficient_block.heighty],
					&coefficient_block, sizeof(QccWAVlcmCoefficientBlock));
			}
		}
	}

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	return(return_value);
}

static int QccWAVlcmPartitionSubsequence(QccWAVSubbandPyramid *subband_pyramid, const QccWAVSubbandPyramid *mask_subband_pyramid, 
	QccWAVlcmCoefficientBlock **subband_lcm_pyramid, int threshold, QccList *nz_neighbors, 
	QccList *nz_parent, QccList *run, QccList *refinement, int *mblock_size, int num_blocks, 
	int mblocks_per_row, int mblocks_per_col, int *nz_neighbors_length, int *nz_parent_length, 
	int *run_length, int *refinement_length)
{
	int return_value;
	int i, j, k, m, n, ii, jj, kk, num_subbands;
	int current_mblock_size;
	int subband_num_rows, subband_num_cols, row_limit, col_limit;
	int row_offset, col_offset, count, temp, shift_count, seq_selected = 0;
	QccListNode *node;
	QccWAVlcmCoefficientBlock **neighbors, *parent;
	int transparent;

	num_subbands = QccWAVSubbandPyramidNumLevelsToNumSubbands(subband_pyramid->num_levels);
	for(k = 1; k < num_subbands; k++) {
		if(QccWAVSubbandPyramidSubbandResolution(subband_pyramid, k, &subband_num_rows, &subband_num_cols)) {
			QccErrorAddMessage("(QccWAVlcmPartitionSubsequence): Error calling QccWAVSubbandPyramidSubbandResolution()");
			goto QccError;
		}
		if(QccWAVSubbandPyramidSubbandOffsets(subband_pyramid, k, &row_offset, &col_offset)) {
			QccErrorAddMessage("(QccWAVlcmPartitionSubsequence): Error calling QccWAVSubbandPyramidSubbandOffsets()");
			goto QccError;
		}
		current_mblock_size = mblock_size[QccWAVSubbandPyramidCalcLevelFromSubband(k, subband_pyramid->num_levels) - 1];
		count = current_mblock_size * current_mblock_size;
		row_limit = row_offset + subband_num_rows; col_limit = col_offset + subband_num_cols;
		for(m = 0; m < mblocks_per_row; m++) {
			for(n = 0; n < mblocks_per_col; n++) {
				for(kk = 0; kk < count; kk++) {
					i = 0; j = 0; temp = 1; shift_count = 0;
					/* interleaving */
					if(kk >= temp) {
						while(kk >= temp) {
							i += (kk & temp) ? (int)pow(2., (shift_count / 2)) : 0;
							temp <<= 1; shift_count++;
							j += (kk & temp) ? (int)pow(2., (shift_count / 2)) : 0;
							temp <<= 1; shift_count++;
						}
					}
					ii = row_offset + m * current_mblock_size + i;
					jj = col_offset + n * current_mblock_size + j;
					if (mask_subband_pyramid == NULL) transparent = 0;
					else transparent = QccAlphaTransparent(mask_subband_pyramid->matrix[ii][jj]);

					if(ii < row_limit && jj < col_limit && (!transparent)) {
						if((node = QccWAVlcmListCreateNode(sizeof(QccWAVlcmCoefficientBlock), (void *)&subband_lcm_pyramid[ii][jj])) == NULL) {
							QccErrorAddMessage("(QccWAVlcmPartitionSubsequence): Error calling QccWAVlcmListCreateNode()");
							goto QccError;                
						}
						seq_selected = 0;
						if(!subband_lcm_pyramid[ii][jj].significant) {
							neighbors = (QccWAVlcmCoefficientBlock **)subband_lcm_pyramid[ii][jj].neighbors;
							if(neighbors != NULL) {
								for(temp = 0; temp < subband_lcm_pyramid[ii][jj].num_neighbors; temp++) {
									if(neighbors[temp]->significant) {
										QccListAppendNode(nz_neighbors, node);
										*nz_neighbors_length += 1;
										seq_selected = 1;
										break;
									}
								}
							}
							if(!seq_selected) {
								parent = (QccWAVlcmCoefficientBlock *)subband_lcm_pyramid[ii][jj].parent;
								if(parent != NULL) {
									if(parent->significant) {
										QccListAppendNode(nz_parent, node);
										*nz_parent_length += 1;
									}
									else {
										QccListAppendNode(run, node);
										*run_length += 1;
									}
								}
								else {
									QccListAppendNode(run, node);
									*run_length += 1;
								}
							}
						}
						else {
							QccListAppendNode(refinement, node);
							*refinement_length += 1;
						}
					}
				}
			}
		}
	}

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	return(return_value);
}

static int QccWAVlcmReadWriteChannel(QccChannel *ch_nz_neighbors, QccChannel *ch_nz_parent, QccChannel *ch_run, 
	QccChannel *ch_refinement, QccList *nz_neighbors, QccList *nz_parent, QccList *run, QccList *refinement, 
	QccChannel *nz_neighbors_sign, QccChannel *nz_parent_sign, QccChannel *run_sign, int **sign_array, int threshold, int method)
{
	int i, k, length, bit, return_value = 0;
	QccWAVlcmCoefficientBlock *coefficient;
	QccListNode *stop, *current_node;

	if(method == QCCWAVLCM_ENCODE) {
		current_node = nz_neighbors->start;
		if(current_node != NULL) {
			stop = nz_neighbors->end->next;
			while(current_node != stop) {
				coefficient = (QccWAVlcmCoefficientBlock *)current_node->value;
				bit = coefficient->value < threshold ? 0 : 1;
				coefficient->significant = bit;        
				if(bit) coefficient->value -= threshold;
				if(QccChannelAddSymbolToChannel(ch_nz_neighbors, bit)) {
					QccErrorAddMessage("(QccWAVlcmReadWriteChannel): Error calling QccChannelAddSymbolToChannel()");
					goto QccError;
				}
				if(bit) {
					if(QccChannelAddSymbolToChannel(nz_neighbors_sign, sign_array[coefficient->widthx][coefficient->heighty])) {
						QccErrorAddMessage("(QccWAVlcmReadWriteChannel): Error calling QccChannelAddSymbolToChannel()");
						goto QccError;
					}
				}
				current_node = current_node->next;
			}
		}

		current_node = nz_parent->start;
		if(current_node != NULL) {
			stop = nz_parent->end->next;
			while(current_node != stop) {
				coefficient = (QccWAVlcmCoefficientBlock *)current_node->value;
				bit = coefficient->value < threshold ? 0 : 1;
				coefficient->significant = bit;        
				if(bit) coefficient->value -= threshold;
				if(QccChannelAddSymbolToChannel(ch_nz_parent, bit)) {
					QccErrorAddMessage("(QccWAVlcmReadWriteChannel): Error calling QccChannelAddSymbolToChannel()");
					goto QccError;
				}
				if(bit) {
					if(QccChannelAddSymbolToChannel(nz_parent_sign, sign_array[coefficient->widthx][coefficient->heighty])) {
						QccErrorAddMessage("(QccWAVlcmReadWriteChannel): Error calling QccChannelAddSymbolToChannel()");
						goto QccError;
					}
				}
				current_node = current_node->next;
			}
		}

		current_node = run->start;
		if(current_node != NULL) {
			stop = run->end->next;
			while(current_node != stop) {
				coefficient = (QccWAVlcmCoefficientBlock *)current_node->value;
				bit = coefficient->value < threshold ? 0 : 1;
				coefficient->significant = bit;        
				if(bit) coefficient->value -= threshold;
				if(QccChannelAddSymbolToChannel(ch_run, bit)) {
					QccErrorAddMessage("(QccWAVlcmReadWriteChannel): Error calling QccChannelAddSymbolToChannel()");
					goto QccError;
				}
				if(bit) {
					if(QccChannelAddSymbolToChannel(run_sign, sign_array[coefficient->widthx][coefficient->heighty])) {
						QccErrorAddMessage("(QccWAVlcmReadWriteChannel): Error calling QccChannelAddSymbolToChannel()");
						goto QccError;
					}
				}
				current_node = current_node->next;
			}
		}

		current_node = refinement->start;
		if(current_node != NULL) {
			stop = refinement->end->next;
			while(current_node != stop) {
				coefficient = (QccWAVlcmCoefficientBlock *)current_node->value;
				bit = coefficient->value < threshold ? 0 : 1;
				if(bit) coefficient->value -= threshold;
				if(QccChannelAddSymbolToChannel(ch_refinement, bit)) {
					QccErrorAddMessage("(QccWAVlcmReadWriteChannel): Error calling QccChannelAddSymbolToChannel()");
					goto QccError;
				}
				current_node = current_node->next;
			}
		}
	}
	else {
		current_node = nz_neighbors->start;
		length = ch_nz_neighbors->channel_length; i = 0; k = 0;
		if(current_node != NULL) {
			stop = nz_neighbors->end->next;
			while(current_node != stop && i < length) {
				bit = ch_nz_neighbors->channel_symbols[i];
				coefficient = (QccWAVlcmCoefficientBlock *)current_node->value;
				coefficient->significant = bit;        
				if(bit) {
					coefficient->value = 1.5 * threshold;
					sign_array[coefficient->widthx][coefficient->heighty] = nz_neighbors_sign->channel_symbols[k++];
				}
				current_node = current_node->next; i++;
			}
			if(i > length) {
				QccErrorAddMessage("(QccWAVlcmReadWriteChannel): Error reading non-zero neighbors channel");
				goto QccError;
			}
		}

		current_node = nz_parent->start;
		length = ch_nz_parent->channel_length; i = 0; k = 0;
		if(current_node != NULL) {
			stop = nz_parent->end->next;
			while(current_node != stop && i < length) {
				bit = ch_nz_parent->channel_symbols[i];
				coefficient = (QccWAVlcmCoefficientBlock *)current_node->value;
				coefficient->significant = bit;        
				if(bit) {
					coefficient->value = 1.5 * threshold;
					sign_array[coefficient->widthx][coefficient->heighty] = nz_parent_sign->channel_symbols[k++];                    
				}
				current_node = current_node->next; i++;
			}
			if(i > length) {
				QccErrorAddMessage("(QccWAVlcmReadWriteChannel): Error reading non-zero parent channel");
				goto QccError;
			}
		}

		current_node = run->start;
		length = ch_run->channel_length; i = 0; k = 0;
		if(current_node != NULL) {
			stop = run->end->next;
			while(current_node != stop && i < length) {
				bit = ch_run->channel_symbols[i];
				coefficient = (QccWAVlcmCoefficientBlock *)current_node->value;
				coefficient->significant = bit;        
				if(bit) {
					coefficient->value = 1.5 * threshold;
					sign_array[coefficient->widthx][coefficient->heighty] = run_sign->channel_symbols[k++];              
				}
				current_node = current_node->next; i++;
			}
			if(i > length) {
				QccErrorAddMessage("(QccWAVlcmReadWriteChannel): Error reading run channel");
				goto QccError;
			}
		}

		current_node = refinement->start;
		length = ch_refinement->channel_length; i = 0;        
		if(current_node != NULL) {
			stop = refinement->end->next;
			while(current_node != stop && i < length) {
				bit = ch_refinement->channel_symbols[i];
				coefficient = (QccWAVlcmCoefficientBlock *)current_node->value;
				coefficient->value += (bit) ? (threshold / 2) : (-threshold / 2);
				current_node = current_node->next; i++;
			}
			if(i > length) {
				QccErrorAddMessage("(QccWAVlcmReadWriteChannel): Error reading refinement channel");
				goto QccError;
			}
		}
	}

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	return(return_value);
}

static int QccWAVlcmEncodeChannel(QccBitBuffer *buffer, QccChannel *nz_neighbors, QccChannel *nz_parent, QccChannel *run, 
	QccChannel *refinement, int target_bit_cnt, int *nz_neighbors_length, int *nz_parent_length, int *run_length, 
	int *refinement_length, QccChannel *nz_neighbors_sign, QccChannel *nz_parent_sign, QccChannel *run_sign)
{
	int i, count, return_value = 0;

	if(nz_neighbors->channel_length != *nz_neighbors_length) {
		QccErrorAddMessage("(QccWAVlcmEncodeChannel): NonZero Neighbors Channel size not match");
		goto QccError;    
	}
	if(QccWAVlcmAdaptiveGolombEncode(buffer, nz_neighbors->channel_symbols, nz_neighbors->channel_length, 
		nz_neighbors_sign, QCCWAVLCM_KMAX)) {
		QccErrorAddMessage("(QccWAVlcmEncodeChannel): Error calling QccWAVlcmAdaptiveGolombEncode()");
		goto QccError;
	}
	if(buffer->bit_cnt >= target_bit_cnt) {
		QccBitBufferFlush(buffer);
		return(2);
	}

	if(nz_parent->channel_length != *nz_parent_length) {
		QccErrorAddMessage("(QccWAVlcmEncodeChannel): NonZero Parent Channel size not match");
		goto QccError;    
	}
	if(QccWAVlcmAdaptiveGolombEncode(buffer, nz_parent->channel_symbols, nz_parent->channel_length, nz_parent_sign, QCCWAVLCM_KMAX)) {
		QccErrorAddMessage("(QccWAVlcmEncodeChannel): Error calling QccWAVlcmAdaptiveGolombEncode()");
		goto QccError;
	}
	if(buffer->bit_cnt >= target_bit_cnt) {
		QccBitBufferFlush(buffer);
		return(2);
	}

	if(run->channel_length != *run_length) {
		QccErrorAddMessage("(QccWAVlcmEncodeChannel): Run Channel size not match");
		goto QccError;    
	}
	if(QccWAVlcmAdaptiveGolombEncode(buffer, run->channel_symbols, run->channel_length, run_sign, QCCWAVLCM_KMAX)) {
		QccErrorAddMessage("(QccWAVlcmEncodeChannel): Error calling QccWAVlcmAdaptiveGolombEncode()");
		goto QccError;
	}
	if(buffer->bit_cnt >= target_bit_cnt) {
		QccBitBufferFlush(buffer);
		return(2);
	}

	if(refinement->channel_length != *refinement_length) {
		QccErrorAddMessage("(QccWAVlcmEncodeChannel): NonZero Neighbors Channel size not match");
		goto QccError;    
	}
	if(refinement->access_block_size != QCCCHANNEL_ACCESSWHOLEFILE) {
		QccErrorAddMessage("(QccWAVlcmEncodeChannel): Refinement channel should be accessed as whole file");
		goto QccError;
	}

	count = refinement->channel_length;    
	for(i = 0; i < count; i++) {
		if(QccWAVlcmInputOutput(buffer, &refinement->channel_symbols[i], QCCWAVLCM_ENCODE, target_bit_cnt)) {
			QccErrorAddMessage("(QccWAVlcmEncodeChannel): Error calling QccWAVlcmInputOutput()");
			return(2);
		}
	}

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	return(return_value);    
}

int QccWAVlcmEncode(const QccIMGImageComponent *image, const QccIMGImageComponent *mask, QccBitBuffer *buffer,
	int num_levels, const QccWAVWavelet *wavelet, const QccWAVPerceptualWeights *perceptual_weights,
	int target_bit_cnt)
{
	int return_value = 0, row;
	unsigned char LL_max_bits;
	QccWAVSubbandPyramid image_subband_pyramid;
	QccWAVSubbandPyramid mask_subband_pyramid;
	QccChannel ch_nz_neighbors, ch_nz_parent, ch_run, ch_refinement;
	QccChannel nz_neighbors_sign, nz_parent_sign, run_sign;
	int **sign_array = NULL;
	double image_mean;
	int max_coefficient_bits;
	QccList nz_neighbors, nz_parent, run, refinement;
	QccWAVlcmCoefficientBlock **subband_lcm_pyramid;
	double threshold;
	int *macro_block_size, num_blocks, macro_block_per_row, macro_block_per_col;
	int nz_neighbors_length, nz_parent_length, run_length, refinement_length;

	if(image == NULL) return(0);
	if(buffer == NULL) return(0);
	if(wavelet == NULL) return(0);

	QccWAVSubbandPyramidInitialize(&image_subband_pyramid);
	QccWAVSubbandPyramidInitialize(&mask_subband_pyramid);

	if((macro_block_size = (int *)malloc(sizeof(int) * num_levels)) == NULL) {
		QccErrorAddMessage("(QccWAVlcmEncode): Error allocating memory");
		goto QccError;        
	}

	QccWAVlcmMacroBlockInit(macro_block_size, &num_blocks, num_levels, image->num_rows, image->num_cols, 
		&macro_block_per_row, &macro_block_per_col);

	image_subband_pyramid.num_levels = 0;
	image_subband_pyramid.num_rows = image->num_rows;
	image_subband_pyramid.num_cols = image->num_cols;
	if(QccWAVSubbandPyramidAlloc(&image_subband_pyramid)) {
		QccErrorAddMessage("(QccWAVlcmEncode): Error calling QccWAVSubbandPyramidAlloc()");
		goto QccError;
	}

	if(mask != NULL) {
		if((mask->num_rows != image->num_rows) || (mask->num_cols != image->num_cols)) {
			QccErrorAddMessage("(QccWAVlcmEncode): Mask and image must be same size");
			goto QccError;
		}

		mask_subband_pyramid.num_levels = 0;
		mask_subband_pyramid.num_rows = mask->num_rows;
		mask_subband_pyramid.num_cols = mask->num_cols;
		if(QccWAVSubbandPyramidAlloc(&mask_subband_pyramid)) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error calling QccWAVSubbandPyramidAlloc()");
			goto QccError;
		}
	}

	if((sign_array = (int **)malloc(sizeof(int *) * image->num_rows)) == NULL) {
		QccErrorAddMessage("(QccWAVlcmEncode): Error allocating memory");
		goto QccError;
	}

	for(row = 0; row < image->num_rows; row++) {
		if((sign_array[row] = (int *)malloc(sizeof(int) * image->num_cols)) == NULL) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error allocating memory");
			goto QccError;
		}
	}

	if((subband_lcm_pyramid = (QccWAVlcmCoefficientBlock **)malloc(sizeof(QccWAVlcmCoefficientBlock *) * image->num_rows)) == NULL) {
		QccErrorAddMessage("(QccWAVlcmEncode): Error allocating memory");
		goto QccError;
	}

	for(row = 0; row < image->num_rows; row++) {
		if((subband_lcm_pyramid[row] = (QccWAVlcmCoefficientBlock *)malloc(sizeof(QccWAVlcmCoefficientBlock) * image->num_cols)) == NULL) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error allocating memory");
			goto QccError;
		}
	}

	if(QccWAVlcmEncodeDWT(&image_subband_pyramid, sign_array, image, num_levels, &image_mean,
		&max_coefficient_bits, &mask_subband_pyramid, mask, wavelet, perceptual_weights)) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error calling QccWAVlcmEncodeDWT()");
			goto QccError;
	}

	if(QccWAVlcmEncodeHeader(buffer, num_levels, image->num_rows, image->num_cols, image_mean,
		max_coefficient_bits)) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error calling QccWAVlcmEncodeHeader()");
			goto QccError;
	}

	if(QccWAVlcmTagInitialize(&image_subband_pyramid, ((mask != NULL) ? &mask_subband_pyramid : NULL), 
		sign_array, num_levels, subband_lcm_pyramid)) {
		QccErrorAddMessage("(QccWAVlcmEncode): Error calling QccWAVlcmTagInitialize()");
		goto QccError;
	}

	threshold = pow((double)2, (double)max_coefficient_bits);
	/* fprintf(stderr, "The value of the threshold is: %d\n", (int)threshold); */
	/* num_subbands = QccWAVSubbandPyramidNumLevelsToNumSubbands(num_levels); */

	/* After calling this function, we have modified sign array for LL subband */
	/* if(QccWAVlcmInputOutputLLMean(buffer, &image_subband_pyramid, subband_lcm_pyramid, sign_array, &LL_max_bits, QCCWAVLCM_ENCODE)) {
		QccErrorAddMessage("(QccWAVlcmEncode): Error calling QccWAVlcmInputOutputLLMean()");
		goto QccError;
	} */

	if(QccWAVlcmInputOutputLLThreshold(buffer, &image_subband_pyramid, &LL_max_bits, QCCWAVLCM_ENCODE)) {
		QccErrorAddMessage("(QccWAVlcmEncode): Error calling QccWAVlcmInputOutputLLThreshold()");
		goto QccError;    
	}

	while(1) {
		QccListInitialize(&nz_neighbors);
		QccListInitialize(&nz_parent);
		QccListInitialize(&run);
		QccListInitialize(&refinement);

		if(QccChannelInitialize(&ch_nz_neighbors)) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error initializing channel");
			goto QccError;
		}
		ch_nz_neighbors.alphabet_size = QCCWAVLCM_ALPHABET_SIZE;

		if(QccChannelInitialize(&ch_nz_parent)) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error initializing channel");
			goto QccError;
		}
		ch_nz_parent.alphabet_size = QCCWAVLCM_ALPHABET_SIZE;

		if(QccChannelInitialize(&ch_run)) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error initializing channel");
			goto QccError;
		}
		ch_run.alphabet_size = QCCWAVLCM_ALPHABET_SIZE;

		if(QccChannelInitialize(&ch_refinement)) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error initializing channel");
			goto QccError;
		}
		ch_refinement.alphabet_size = QCCWAVLCM_ALPHABET_SIZE;

		if(QccChannelInitialize(&nz_neighbors_sign)) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error initializing channel");
			goto QccError;
		}

		if(QccChannelInitialize(&nz_parent_sign)) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error initializing channel");
			goto QccError;
		}

		if(QccChannelInitialize(&run_sign)) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error initializing channel");
			goto QccError;
		}

		return_value = QccWAVlcmInputOutputLLBitplane(buffer, &image_subband_pyramid, subband_lcm_pyramid, sign_array, 
			threshold, (int)LL_max_bits, target_bit_cnt, QCCWAVLCM_ENCODE);
		if(return_value == 1){
			QccErrorAddMessage("(QccWAVlcmEncode): Error calling QccWAVlcmInputOutputLLBitplane()");
			goto QccError;
		}
		else if(return_value == 2) break;

		nz_neighbors_length = 0; nz_parent_length = 0;
		run_length = 0; refinement_length = 0;
		if(QccWAVlcmPartitionSubsequence(&image_subband_pyramid, ((mask != NULL) ? &mask_subband_pyramid : NULL), 
			subband_lcm_pyramid, threshold, &nz_neighbors, &nz_parent, &run, &refinement, macro_block_size, 
			num_blocks, macro_block_per_row, macro_block_per_col, &nz_neighbors_length, &nz_parent_length, 
			&run_length, &refinement_length)) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error calling QccWAVlcmPartitionSubsequence()");
			goto QccError;
		}

		if(QccWAVlcmReadWriteChannel(&ch_nz_neighbors, &ch_nz_parent, &ch_run, &ch_refinement, &nz_neighbors,
			&nz_parent, &run, &refinement, &nz_neighbors_sign, &nz_parent_sign, &run_sign, sign_array, threshold, QCCWAVLCM_ENCODE)) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error calling QccWAVlcmReadWriteChannel()");
			goto QccError;        
		}

		return_value = QccWAVlcmEncodeChannel(buffer, &ch_nz_neighbors, &ch_nz_parent, &ch_run, &ch_refinement, target_bit_cnt, 
			&nz_neighbors_length, &nz_parent_length, &run_length, &refinement_length, &nz_neighbors_sign, &nz_parent_sign,
			&run_sign); 
		if(return_value == 1) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error calling QccWAVlcmEncodeChannel()");
			goto QccError;        
		}
		else if(return_value == 2) break;

		threshold /= 2.0;
		if(threshold < 1.0) break;

		QccWAVlcmListFree(&nz_neighbors); QccWAVlcmListFree(&nz_parent); 
		QccWAVlcmListFree(&run); QccWAVlcmListFree(&refinement);
		QccChannelFree(&ch_nz_neighbors); QccChannelFree(&ch_nz_parent);
		QccChannelFree(&ch_run); QccChannelFree(&ch_refinement);
		QccChannelFree(&nz_neighbors_sign); QccChannelFree(&nz_parent_sign);
		QccChannelFree(&run_sign);
	}

	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	QccWAVSubbandPyramidFree(&image_subband_pyramid);
	QccWAVSubbandPyramidFree(&mask_subband_pyramid);

	if(sign_array != NULL) {
		for (row = 0; row < image->num_rows; row++) {
			if (sign_array[row] != NULL) free(sign_array[row]);
		}
		free(sign_array);
	}

	if(subband_lcm_pyramid != NULL) {
		for (row = 0; row < image->num_rows; row++) {
			if (subband_lcm_pyramid[row] != NULL) free(subband_lcm_pyramid[row]);
		}
		free(subband_lcm_pyramid);
	}

	if(return_value) {
		QccWAVlcmListFree(&nz_neighbors); QccWAVlcmListFree(&nz_parent); 
		QccWAVlcmListFree(&run); QccWAVlcmListFree(&refinement);

		QccChannelFree(&ch_nz_neighbors); QccChannelFree(&ch_nz_parent);
		QccChannelFree(&ch_run); QccChannelFree(&ch_refinement);
		QccChannelFree(&nz_neighbors_sign); QccChannelFree(&nz_parent_sign);
		QccChannelFree(&run_sign);        
	}
	if(return_value == 2) return_value = 0;
	return(return_value);
}

int QccWAVlcmDecodeHeader(QccBitBuffer *buffer, int *num_levels, int *num_rows, int *num_cols,
	double *image_mean, int *max_coefficient_bits)
{
	int return_value;
	unsigned char ch;

	if(QccBitBufferGetChar(buffer, &ch)) {
		QccErrorAddMessage("(QccWAVlcmDecodeHeader): Error calling QccBitBufferGetChar()");
		goto QccError;
	}
	*num_levels = (int)ch;

	if(QccBitBufferGetInt(buffer, num_rows)) {
		QccErrorAddMessage("(QccWAVlcmDecodeHeader): Error calling QccBitBufferGetInt()");
		goto QccError;
	}

	if(QccBitBufferGetInt(buffer, num_cols)) {
		QccErrorAddMessage("(QccWAVlcmDecodeHeader): Error calling QccBitBufferGetInt()");
		goto QccError;
	}

	if(QccBitBufferGetDouble(buffer, image_mean)) {
		QccErrorAddMessage("(QccWAVlcmDecodeHeader): Error calling QccBitBufferGetDouble()");
		goto QccError;
	}

	if(QccBitBufferGetChar(buffer, &ch)) {
		QccErrorAddMessage("(QccWAVlcmDecodeHeader): Error calling QccBitBufferGetChar()");
		goto QccError;
	}
	*max_coefficient_bits = (int)ch;

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	return(return_value);
}

static int QccWAVlcmDecodeChannel(QccBitBuffer *buffer, QccChannel *nz_neighbors, QccChannel *nz_parent, QccChannel *run, 
	QccChannel *refinement, int *nz_neighbors_length, int *nz_parent_length, int *run_length, int *refinement_length,
	QccChannel *nz_neighbors_sign, QccChannel *nz_parent_sign, QccChannel *run_sign)
{
	int i, count, return_value;
	int target_bit_cnt = 0;

	nz_neighbors->channel_length = *nz_neighbors_length;
	if(QccChannelAlloc(nz_neighbors)) {
		QccErrorAddMessage("(QccWAVlcmDecodeChannel): Error calling QccChannelAlloc()");
		goto QccError;    
	}
	if(QccWAVlcmAdaptiveGolombDecode(buffer, nz_neighbors->channel_symbols, *nz_neighbors_length, 
		nz_neighbors_sign, QCCWAVLCM_KMAX)) {
		QccErrorAddMessage("(QccWAVlcmDecodeChannel): Error calling QccWAVlcmAdaptiveGolombDecode()");
		goto QccError;
	}

	nz_parent->channel_length = *nz_parent_length;
	if(QccChannelAlloc(nz_parent)) {
		QccErrorAddMessage("(QccWAVlcmDecodeChannel): Error calling QccChannelAlloc()");
		goto QccError;    
	}
	if(QccWAVlcmAdaptiveGolombDecode(buffer, nz_parent->channel_symbols, *nz_parent_length, nz_parent_sign, QCCWAVLCM_KMAX)) {
		QccErrorAddMessage("(QccWAVlcmDecodeChannel): Error calling QccWAVlcmAdaptiveGolombDecode()");
		goto QccError;
	}

	run->channel_length = *run_length;
	if(QccChannelAlloc(run)) {
		QccErrorAddMessage("(QccWAVlcmDecodeChannel): Error calling QccChannelAlloc()");
		goto QccError;    
	}
	if(QccWAVlcmAdaptiveGolombDecode(buffer, run->channel_symbols, *run_length, run_sign, QCCWAVLCM_KMAX)) {
		QccErrorAddMessage("(QccWAVlcmDecodeChannel): Error calling QccWAVlcmAdaptiveGolombDecode()");
		goto QccError;
	}

	refinement->channel_length = *refinement_length;
	if(QccChannelAlloc(refinement)) {
		QccErrorAddMessage("(QccWAVlcmDecodeChannel): Error calling QccChannelAlloc()");
		goto QccError;    
	}
	if(refinement->access_block_size != QCCCHANNEL_ACCESSWHOLEFILE) {
		QccErrorAddMessage("(QccWAVlcmDecodeChannel): Refinement channel should be accessed as whole file");
		goto QccError;
	}

	count = *refinement_length;    
	for(i = 0; i < count; i++) {
		if(QccWAVlcmInputOutput(buffer, &refinement->channel_symbols[i], QCCWAVLCM_DECODE, target_bit_cnt)) {
			QccErrorAddMessage("(QccWAVlcmDecodeChannel): Error calling QccWAVlcmInputOutput()");
			return(2);
		}
	}

	return_value = 0;
	goto QccReturn;
QccError:
	return_value = 2;
QccReturn:
	return(return_value);    
}

int QccWAVlcmDecode(QccBitBuffer *buffer, QccIMGImageComponent *image, const QccIMGImageComponent *mask,
	int num_levels, const QccWAVWavelet *wavelet, const QccWAVPerceptualWeights *perceptual_weights,
	double image_mean, int max_coefficient_bits)
{
	int return_value = 0, target_bit_cnt = 0;
	unsigned char LL_max_bits;
	QccWAVSubbandPyramid image_subband_pyramid;
	QccWAVSubbandPyramid mask_subband_pyramid;
	int **sign_array = NULL;
	QccList nz_neighbors, nz_parent, run, refinement;
	double threshold;
	int i, j, row, col;
	QccWAVWavelet lazy_wavelet_transform;
	QccChannel ch_nz_neighbors, ch_nz_parent, ch_run, ch_refinement;
	QccChannel nz_neighbors_sign, nz_parent_sign, run_sign;
	int nz_neighbors_length, nz_parent_length, run_length, refinement_length;
	QccWAVlcmCoefficientBlock **subband_lcm_pyramid;
	int *macro_block_size, num_blocks, macro_block_per_row, macro_block_per_col;

	if(image == NULL) return(0);
	if(buffer == NULL) return(0);
	if(wavelet == NULL) return(0);

	QccWAVSubbandPyramidInitialize(&image_subband_pyramid);
	QccWAVSubbandPyramidInitialize(&mask_subband_pyramid);

	if((macro_block_size = (int *)malloc(sizeof(int) * num_levels)) == NULL) {
		QccErrorAddMessage("(QccWAVlcmEncode): Error allocating memory");
		goto QccError;        
	}

	QccWAVlcmMacroBlockInit(macro_block_size, &num_blocks, num_levels, image->num_rows, image->num_cols, 
		&macro_block_per_row, &macro_block_per_col);

	QccWAVWaveletInitialize(&lazy_wavelet_transform);

	image_subband_pyramid.num_levels = num_levels;
	image_subband_pyramid.num_rows = image->num_rows;
	image_subband_pyramid.num_cols = image->num_cols;
	if(QccWAVSubbandPyramidAlloc(&image_subband_pyramid)) {
		QccErrorAddMessage("(QccWAVlcmDecode): Error calling QccWAVSubbandPyramidAlloc()");
		goto QccError;
	}

	if(mask != NULL) {
		if((mask->num_rows != image->num_rows) || (mask->num_cols != image->num_cols)) {
			QccErrorAddMessage("(QccWAVlcmDecode): Mask and image must be same size");
			goto QccError;
		}

		if(QccWAVWaveletCreate(&lazy_wavelet_transform, "LWT.lft", "symmetric")) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error calling QccWAVWaveletCreate()");
			goto QccError;
		}

		mask_subband_pyramid.num_levels = 0;
		mask_subband_pyramid.num_rows = mask->num_rows;
		mask_subband_pyramid.num_cols = mask->num_cols;
		if(QccWAVSubbandPyramidAlloc(&mask_subband_pyramid)) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error calling QccWAVSubbandPyramidAlloc()");
			goto QccError;
		}

		if(QccMatrixCopy(mask_subband_pyramid.matrix, mask->image, mask->num_rows, mask->num_cols)) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error calling QccMatrixCopy()");
			goto QccError;
		}

		if(QccWAVSubbandPyramidDWT(&mask_subband_pyramid, num_levels, &lazy_wavelet_transform)) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error calling QccWAVSubbandPyramidDWT()");
			goto QccError;
		}
	}
 
	if((sign_array = (int **)malloc(sizeof(int *) * image->num_rows)) == NULL) {
		QccErrorAddMessage("(QccWAVlcmDecode): Error allocating memory");
		goto QccError;
	}

	for(row = 0; row < image->num_rows; row++) {
		if((sign_array[row] = (int *)malloc(sizeof(int) * image->num_cols)) == NULL) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error allocating memory");
			goto QccError;
		}
	}

	for(row = 0; row < image->num_rows; row++) {
		for (col = 0; col < image->num_cols; col++) {
			image_subband_pyramid.matrix[row][col] = 0.0;
			sign_array[row][col] = 0;
		}
	}

	if((subband_lcm_pyramid = (QccWAVlcmCoefficientBlock **)malloc(sizeof(QccWAVlcmCoefficientBlock *) * image->num_rows)) == NULL) {
		QccErrorAddMessage("(QccWAVlcmDecode): Error allocating memory");
		goto QccError;
	}

	for(row = 0; row < image->num_rows; row++) {
		if((subband_lcm_pyramid[row] = (QccWAVlcmCoefficientBlock *)malloc(sizeof(QccWAVlcmCoefficientBlock) * image->num_cols)) == NULL) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error allocating memory");
			goto QccError;
		}
	}

	if(QccWAVlcmTagInitialize(&image_subband_pyramid, ((mask != NULL) ? &mask_subband_pyramid : NULL), 
		sign_array, num_levels, subband_lcm_pyramid)) {
		QccErrorAddMessage("(QccWAVlcmDecode): Error calling QccWAVlcmTagInitialize()");
		goto QccError;
	}

	threshold = pow((double)2, (double)max_coefficient_bits);

	/* if(QccWAVlcmInputOutputLLMean(buffer, &image_subband_pyramid, subband_lcm_pyramid, sign_array, &LL_max_bits, QCCWAVLCM_DECODE)) {
		QccErrorAddMessage("(QccWAVlcmDecode): Error calling QccWAVlcmInputOutputLLMean()");
		goto QccError;
	} */
	if(QccWAVlcmInputOutputLLThreshold(buffer, &image_subband_pyramid, &LL_max_bits, QCCWAVLCM_DECODE)) {
		QccErrorAddMessage("(QccWAVlcmDecode): Error calling QccWAVlcmInputOutputLLThreshold()");
		goto QccError;    
	}

	while(1) {
		QccListInitialize(&nz_neighbors);
		QccListInitialize(&nz_parent);
		QccListInitialize(&run);
		QccListInitialize(&refinement);

		if(QccChannelInitialize(&ch_nz_neighbors)) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error initializing channel");
			goto QccError;
		}
		ch_nz_neighbors.alphabet_size = QCCWAVLCM_ALPHABET_SIZE;

		if(QccChannelInitialize(&ch_nz_parent)) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error initializing channel");
			goto QccError;
		}
		ch_nz_parent.alphabet_size = QCCWAVLCM_ALPHABET_SIZE;

		if(QccChannelInitialize(&ch_run)) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error initializing channel");
			goto QccError;
		}
		ch_run.alphabet_size = QCCWAVLCM_ALPHABET_SIZE;

		if(QccChannelInitialize(&ch_refinement)) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error initializing channel");
			goto QccError;
		}
		ch_refinement.alphabet_size = QCCWAVLCM_ALPHABET_SIZE;

		if(QccChannelInitialize(&nz_neighbors_sign)) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error initializing channel");
			goto QccError;
		}

		if(QccChannelInitialize(&nz_parent_sign)) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error initializing channel");
			goto QccError;
		}

		if(QccChannelInitialize(&run_sign)) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error initializing channel");
			goto QccError;
		}

		return_value = QccWAVlcmInputOutputLLBitplane(buffer, &image_subband_pyramid, subband_lcm_pyramid, sign_array, 
			threshold, (int)LL_max_bits, target_bit_cnt, QCCWAVLCM_DECODE);
		if(return_value == 1){
			QccErrorAddMessage("(QccWAVlcmDecode): Error calling QccWAVlcmInputOutputLLBitplane()");
			goto QccError;
		}
		else if(return_value == 2) break;

		nz_neighbors_length = 0; nz_parent_length = 0;
		run_length = 0; refinement_length = 0;
		if(QccWAVlcmPartitionSubsequence(&image_subband_pyramid, ((mask != NULL) ? &mask_subband_pyramid : NULL), subband_lcm_pyramid, 
			threshold, &nz_neighbors, &nz_parent, &run, &refinement, macro_block_size, num_blocks, macro_block_per_row, 
			macro_block_per_col, &nz_neighbors_length, &nz_parent_length, &run_length, &refinement_length)) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error calling QccWAVlcmPartitionSubsequence()");
			goto QccError;
		}

		return_value = QccWAVlcmDecodeChannel(buffer, &ch_nz_neighbors, &ch_nz_parent, &ch_run, &ch_refinement,
			&nz_neighbors_length, &nz_parent_length, &run_length, &refinement_length, &nz_neighbors_sign,
			&nz_parent_sign, &run_sign); 
		if(return_value == 1) {
			QccErrorAddMessage("(QccWAVlcmEncode): Error calling QccWAVlcmEncodeChannel()");
			goto QccError;
		}
		else if(return_value == 2) break;

		if(QccWAVlcmReadWriteChannel(&ch_nz_neighbors, &ch_nz_parent, &ch_run, &ch_refinement, &nz_neighbors,
			&nz_parent, &run, &refinement, &nz_neighbors_sign, &nz_parent_sign, &run_sign, sign_array, 
			threshold, QCCWAVLCM_DECODE)) {
			QccErrorAddMessage("(QccWAVlcmDecode): Error calling QccWAVlcmReadWriteChannel()");
			goto QccError;
		}

		threshold /= 2.0;
		if(threshold < (double)1) break;

		QccWAVlcmListFree(&nz_neighbors); QccWAVlcmListFree(&nz_parent); 
		QccWAVlcmListFree(&run); QccWAVlcmListFree(&refinement);
		QccChannelFree(&ch_nz_neighbors); QccChannelFree(&ch_nz_parent);
		QccChannelFree(&ch_run); QccChannelFree(&ch_refinement);
		QccChannelFree(&nz_neighbors_sign); QccChannelFree(&nz_parent_sign);
		QccChannelFree(&run_sign);
	}

	for(i = 0; i < image->num_rows; i++) {
		for(j = 0; j < image->num_cols; j++) image_subband_pyramid.matrix[i][j] = subband_lcm_pyramid[i][j].value;
	}

	if(QccWAVlcmDecodeInverseDWT(&image_subband_pyramid, ((mask != NULL) ? &mask_subband_pyramid : NULL),
		sign_array, image, image_mean, wavelet, perceptual_weights)) {
		QccErrorAddMessage("(QccWAVlcmDecode): Error calling QccWAVlcmDecodeInverseDWT()");
		goto QccError;
	}

	QccErrorClearMessages();
	goto QccReturn;
QccError:
	return_value = 1;
QccReturn:
	QccWAVSubbandPyramidFree(&image_subband_pyramid);
	QccWAVSubbandPyramidFree(&mask_subband_pyramid);
	QccWAVWaveletFree(&lazy_wavelet_transform);
	if(sign_array != NULL) {
		for(row = 0; row < image->num_rows; row++) {
			if(sign_array[row] != NULL) free(sign_array[row]);
		}
		free(sign_array);
	}

	if(subband_lcm_pyramid != NULL) {
		for (row = 0; row < image->num_rows; row++) {
			if (subband_lcm_pyramid[row] != NULL) free(subband_lcm_pyramid[row]);
		}
		free(subband_lcm_pyramid);
	}

	if(return_value) {
		QccWAVlcmListFree(&nz_neighbors); QccWAVlcmListFree(&nz_parent); 
		QccWAVlcmListFree(&run); QccWAVlcmListFree(&refinement);
		QccChannelFree(&ch_nz_neighbors); QccChannelFree(&ch_nz_parent);
		QccChannelFree(&ch_run); QccChannelFree(&ch_refinement);
		QccChannelFree(&nz_neighbors_sign); QccChannelFree(&nz_parent_sign);
		QccChannelFree(&run_sign);        
	}
	if(return_value == 2) return_value = 0;
	return(return_value);
}
