/*
 *  This code was written by Yufei Yuan <yuanyufei@hotmail.com>
 */

#include "lcmdecode.h"

#define USG_STRING "[-w %s:wavelet] [-b %s:boundary] [-pw %: %s:pcpfile] [-m %: %s:mask] %s:bitstream %s:imgfile"

QccWAVWavelet Wavelet;
QccString WaveletFilename = QCCWAVWAVELET_DEFAULT_WAVELET;
QccString Boundary = "symmetric";

QccWAVPerceptualWeights PerceptualWeights;
int UsePerceptualWeights = 0;

int NumLevels;
int NumSubbands;

QccBitBuffer InputBuffer;

int NumRows, NumCols;

QccIMGImage OutputImage;

double ImageMean;
int MaxCoefficientBits;

QccIMGImage Mask;
int MaskSpecified = 0;
int MaskNumRows, MaskNumCols;

int main(int argc, char *argv[])
{
	QccInit(argc, argv);
	QccWAVlcmHeader();

	QccWAVWaveletInitialize(&Wavelet);
	QccWAVPerceptualWeightsInitialize(&PerceptualWeights);
	QccIMGImageInitialize(&OutputImage);
	QccIMGImageInitialize(&Mask);
	QccBitBufferInitialize(&InputBuffer);

	if(QccParseParameters(argc, argv, USG_STRING, WaveletFilename, Boundary,
		&UsePerceptualWeights, PerceptualWeights.filename, &MaskSpecified, Mask.filename,
		InputBuffer.filename, OutputImage.filename))
		QccErrorExit();

	if(QccWAVWaveletCreate(&Wavelet, WaveletFilename, Boundary)) {
		QccErrorAddMessage("%s: Error calling QccWAVWaveletCreate()", argv[0]);
		QccErrorExit();
	}

	InputBuffer.type = QCCBITBUFFER_INPUT;
	if(QccBitBufferStart(&InputBuffer)) {
		QccErrorAddMessage("%s: Error calling QccBitBufferStart()", argv[0]);
		QccErrorExit();
	}

	if(QccWAVlcmDecodeHeader(&InputBuffer, &NumLevels, &NumRows, &NumCols, &ImageMean,
		&MaxCoefficientBits)) {
		QccErrorAddMessage("%s: Error calling QccWAVlcmDecodeHeader()", argv[0]);
		QccErrorExit();
	}
	NumSubbands = QccWAVSubbandPyramidNumLevelsToNumSubbands(NumLevels);

	if(MaskSpecified) {
		if(QccIMGImageRead(&Mask)) {
			QccErrorAddMessage("%s: Error calling QccIMGImageRead()", argv[0]);
			QccErrorExit();
		}
		if(QccIMGImageColor(&Mask)) {
			QccErrorAddMessage("%s: Mask must be grayscale", argv[0]);
			QccErrorExit();
		}
		if(QccIMGImageGetResolution(&Mask, &MaskNumRows, &MaskNumCols)) {
			QccErrorAddMessage("%s: Error calling QccIMGImageGetResolution()", argv[0]);
			QccErrorExit();
		}

		if((MaskNumRows != NumRows) || (MaskNumCols != NumCols)) {
			QccErrorAddMessage("%s: Mask must be same size as image", argv[0]);
			QccErrorExit();
		}
	}

	OutputImage.image_type = QCCIMGTYPE_PGM;
	if(QccIMGImageSetResolution(&OutputImage, NumRows, NumCols)) {
		QccErrorAddMessage("%s: Error calling QccIMGImageSetResolution()", argv[0]);
		QccErrorExit();
	}
	if(QccIMGImageAlloc(&OutputImage)) {
		QccErrorAddMessage("%s: Error calling QccIMGImageAlloc()", argv[0]);
		QccErrorExit();
	}

	if(UsePerceptualWeights) {
		if(QccWAVPerceptualWeightsRead(&PerceptualWeights)) {
			QccErrorAddMessage("%s: Error calling QccWAVPerceptualWeightsRead()", argv[0]);
			QccErrorExit();
		}
		if(PerceptualWeights.num_subbands != NumSubbands) {
			QccErrorAddMessage("%s: Number of subbands in %s does not match that specified in %s",
				argv[0], PerceptualWeights.filename, InputBuffer.filename);
		}
	}

	if(QccWAVlcmDecode(&InputBuffer, &(OutputImage.Y), (MaskSpecified ? &(Mask.Y) : NULL),
		NumLevels, &Wavelet, ((UsePerceptualWeights) ? &PerceptualWeights : NULL),
		ImageMean, MaxCoefficientBits)) {
		QccErrorAddMessage("%s: Error calling QccWAVlcmDecode()", argv[0]);
		QccErrorExit();
	}

	if(QccIMGImageWrite(&OutputImage)) {
		QccErrorAddMessage("%s: Error calling QccIMGImageWrite()", argv[0]);
		QccErrorExit();
	}

	if(QccBitBufferEnd(&InputBuffer)) {
		QccErrorAddMessage("%s: Error calling QccBitBufferEnd()", argv[0]);
		QccErrorExit();
	}

	QccExit;
}
