/*
 *  This code was written by Yufei Yuan <yuanyufei@hotmail.com>
 */

#ifndef LIBQCCPACKLCM_H
#define LIBQCCPACKLCM_H


/*  lcm.c  */
#define QCCWAVLCM_LICENSE "\n***The original algorithm was proposed by E. Ordentlich, M. Weinberger, and G. Seroussi***\n"
#define QCCWAVLCM_MAJORVERSION 0
#define QCCWAVLCM_MINORVERSION 41
#define QCCWAVLCM_COPYRIGHT "Copyright (C) 2001-2003 Yufei Yuan"
#define QCCWAVLCM_DATE "18-april-2003"

#define QCCWAVLCM_NZ_NEIGHBOR_SEQ 0
#define QCCWAVLCM_NZ_PARENT_SEQ 1
#define QCCWAVLCM_RUN_SEQ 2
#define QCCWAVLCM_REFINE_SEQ 3
#define QCCWAVLCM_BLOCK_EXP 7

#define QCCWAVLCM_ENCODE 0
#define QCCWAVLCM_DECODE 1


#define QCCWAVLCM_ALPHABET_SIZE 2
#define QCCWAVLCM_REPEAT 0
#define QCCWAVLCM_STOP 1
#define QCCWAVLCM_KMAX 32


typedef struct
{
	int widthx;
	int heighty;
	void *parent;
	int num_neighbors;
	void *neighbors;
	int value; /*absolute value*/
	unsigned int significant;
} QccWAVlcmCoefficientBlock;

void QccWAVlcmHeader(void);

int QccWAVlcmEncodeHeader(QccBitBuffer *buffer, 
	int num_levels, 
	int num_rows, 
	int num_cols,
	double image_mean, 
	int max_coefficient_bits);

int QccWAVlcmEncode(const QccIMGImageComponent *image,
	const QccIMGImageComponent *mask,
	QccBitBuffer *buffer,
	int num_levels,
	const QccWAVWavelet *wavelet,
	const QccWAVPerceptualWeights *perceptual_weights,
	int target_bit_cnt);

int QccWAVlcmDecodeHeader(QccBitBuffer *buffer,
	int *num_levels,
	int *num_rows, int *num_cols,
	double *image_mean,
	int *max_coefficient_bits);

int QccWAVlcmDecode(QccBitBuffer *buffer,
	QccIMGImageComponent *image,
	const QccIMGImageComponent *mask,
	int num_levels,
	const QccWAVWavelet *wavelet,
	const QccWAVPerceptualWeights *perceptual_weights,
	double image_mean,
	int max_coefficient_bits);

#endif /* LIBQCCPACKLCM_H */
